package fr.username404.snowygui.config

import com.typesafe.config.*
import com.typesafe.config.ConfigFactory.*
import fr.username404.snowygui.ClickGui
import fr.username404.snowygui.Snowy
import fr.username404.snowygui.gui.feature.ButtonImpl
import fr.username404.snowygui.gui.feature.Macro
import fr.username404.snowygui.gui.feature.riskyCheatsEnabled
import io.github.config4k.extract
import kotlinx.coroutines.coroutineScope
import kotlinx.coroutines.launch
import kotlinx.coroutines.runBlocking
import net.minecraft.client.Minecraft
import java.io.File
import java.nio.charset.Charset
import java.util.function.Supplier

object Configuration {
    val ModifiableValues = mutableMapOf<String, Supplier<ConfigValue>>(
        "enabledFeatures" to Supplier { ConfigValueFactory.fromMap(enabledFeatures) },
        "riskyCheats" to Supplier { ConfigValueFactory.fromAnyRef(riskyCheatsEnabled) }
    ); inline fun setConfigValue(s: String, crossinline value: () -> Any) = ModifiableValues.put(s, Supplier { ConfigValueFactory.fromAnyRef(value.invoke()) })
    private fun Config.withFullModifiableValues() = ModifiableValues.entries.fold(this) { previous, entry ->
        previous.withValue(entry.key, entry.value.get())
    }
    private val file: File = File(Minecraft.getInstance().gameDirectory.absolutePath + File.separator + "snowy.conf")
    private val macroFile: File = File(Minecraft.getInstance().gameDirectory.absolutePath + File.separator + "snowy-macros.json")
    private suspend fun writeConfig(c: Config) = coroutineScope {
        launch {
            file.writeText(
                """
                |Snowy {
                |${
                      c.root().render(
                          ConfigRenderOptions.defaults()
                              .setFormatted(true)
                              .setJson(false)
                              .setOriginComments(false)
                      ).prependIndent("  ").trimEnd()
                  }
                |}
                """.trimMargin()
            )
        }
    }
    val obtained: Config by lazy {
        var result: Config = empty()
        with(file) {
            if (!exists()) {
                createNewFile()
                setWritable(true)
                setReadable(true)
            } else try {
                result = parseFile(file)
            } catch (e: ConfigException) {
                Snowy.logs.warn("Could not parse the snowy configuration file, the default configuration will be used instead.")
            }
        }
        load(result).withFallback(
            parseString(
                """
                    |Snowy {
                    |  displayInitMessage = false
                    |  riskyCheats = false
                    |}
                """.trimMargin()
            )
        ).extract<Config>("Snowy")
    }.also {
        runBlocking {
            writeConfig(it.value)
        }
    }
    val enabledFeatures = mutableMapOf<String, Boolean>().apply {
        if (obtained.hasPath("enabledFeatures")) {
            putAll(obtained.extract("enabledFeatures"))
        }
    }
    val macros: MutableSet<Macro> = run {
        if (!macroFile.exists()) macroFile.createNewFile()
        macroFile.readLines(Charset.forName("UTF-8")).filterNot { it.isBlank() }.map {
            Snowy.Gson.fromJson(it, Macro::class.java).copy()
        }.toMutableSet()
    }
    init {
        Runtime.getRuntime().addShutdownHook(
            Thread {
                runBlocking {
                    ClickGui.boxContext {
                        enabledFeatures.putAll(
                            buttons.map {
                                (it as ButtonImpl).run {
                                    title to toggled
                                }
                            }
                        )
                    }
                    macros.map { Snowy.Gson.toJson(it) }.forEachIndexed { index, s ->
                        with(macroFile) {
                            "$s\n".let {
                                if (index == 0) {
                                    writeText(it)
                                } else appendText(it)
                            }
                        }
                    }
                    writeConfig(obtained.withFullModifiableValues()).join()
                }
            }
        )
    }
}
